import { SerdeOptions } from '../option'
import { isFunction, isString, IterResult, Nullable } from '../utils'
import { Deserialize, IDeserializer, IIterableAccess, IMapAccess, IVisitor } from './interface'

export class GenericSeed<T> {
  readonly visitor: IVisitor<T>

  constructor(visitor: IVisitor<T> = new ProxyVisitor()) {
    this.visitor = visitor
  }

  static deserialize<T, D extends IDeserializer>(deserializer: D, visitor: IVisitor<T> = new ProxyVisitor()): T {
    return deserializer.deserializeAny(visitor)
  }

  deserialize<D extends IDeserializer>(deserializer: D): T {
    return GenericSeed.deserialize(deserializer, this.visitor)
  }
}

class ProxyMapAccess implements IMapAccess {
  private access!: IMapAccess
  private options?: SerdeOptions

  constructor(access: IMapAccess, options?: SerdeOptions) {
    if (access instanceof ProxyMapAccess) {
      return this
    }

    this.access = access
    this.options = options
  }

  private wrapResponse<T, I extends IteratorResult<T> = IteratorResult<T>>(result: I): I {
    if (result.done) {
      return result
    } else if (isString(result.value)) {
      const key = this.options?.getDeserializationPropertyName(result.value) ?? result.value
      return IterResult.Next(key) as I
    } else if (Array.isArray(result.value)) {
      const [alias, value] = result.value
      const key = this.options?.getDeserializationPropertyName(alias) ?? alias
      return IterResult.Next([
        key,
        value
      ]) as I
    } else {
      return result
    }
  }

  nextKeySeed<T, K extends Deserialize<T>>(seed: K): IteratorResult<T> {
    return this.wrapResponse<T>(this.access.nextKeySeed(seed))
  }

  nextValueSeed<T, V extends Deserialize<T>>(seed: V): IteratorResult<T> {
    return this.access.nextValueSeed(seed)
  }

  nextEntrySeed<TK, TV, K extends Deserialize<TK>, V extends Deserialize<TV>>(kseed: K, vseed: V): IteratorResult<[TK, TV]> {
    return this.wrapResponse<[TK, TV]>(this.access.nextEntrySeed(kseed, vseed))
  }

  nextKey<T>(): IteratorResult<T> {
    return this.wrapResponse<T>(this.access.nextKey())
  }

  nextValue<V>(): IteratorResult<V> {
    return this.access.nextValue()
  }

  nextEntry<K, V>(): IteratorResult<[K, V]> {
    return this.wrapResponse<[K, V]>(this.access.nextEntry())
  }

  sizeHint?(): Nullable<number> {
    return this.access.sizeHint?.()
  }

  private *generate<T>(next: () => IteratorResult<T>): Iterator<T> {
    let item: IteratorResult<T>

    while ((item = next()) && !item.done) {
      yield item.value
    }
  }

  keys<T, K extends Deserialize<T>>(seed?: K): Iterator<T> {
    return this.generate(
      seed == null ?
        this.nextKey.bind(this) :
        this.nextKeySeed.bind(this, seed) as any
    )
  }

  values<T, V extends Deserialize<T>>(seed?: V): Iterator<T> {
    return this.generate(
      seed == null ?
        this.nextValue.bind(this) :
        this.nextValueSeed.bind(this, seed) as any
    )
  }

  entries<TK, TV, K extends Deserialize<TK>, V extends Deserialize<TV>>(kseed?: K, vseed?: V): Iterator<[TK, TV]> {
    return this.generate(
      kseed == null || vseed == null ?
        this.nextEntry.bind(this) :
        this.nextEntrySeed.bind(this, kseed, vseed) as any
    )
  }

  [Symbol.iterator]<K, V>(): Iterator<[K, V]> {
    return this.entries() as Iterator<[K, V]>
  }
}

export class ProxyVisitor<T> implements IVisitor<T> {
  private overrides?: Partial<IVisitor<T>>
  private options?: SerdeOptions

  constructor(overrides?: Partial<IVisitor<T>>, options?: SerdeOptions) {
    if (overrides instanceof ProxyVisitor) {
      return overrides
    }

    this.overrides = overrides
    this.options = options
  }

  visitBoolean(value: boolean): T {
    return this.overrides?.visitBoolean?.(value) ?? value as T
  }

  visitNumber(value: number): T {
    return this.overrides?.visitNumber?.(value) ?? value as T
  }

  visitBigInt(value: bigint): T {
    return this.overrides?.visitBigInt?.(value) ?? value as T
  }

  visitString(value: string): T {
    return this.overrides?.visitString?.(value) ?? value as T
  }

  visitSymbol(value: symbol): T {
    return this.overrides?.visitSymbol?.(value) ?? value as T
  }

  visitNull(): T {
    return this.overrides?.visitNull?.() ?? null as T
  }

  visitObject(access: IMapAccess): T {
    const proxy = new ProxyMapAccess(access, this.options)

    if (isFunction(this.overrides?.visitObject)) {
      return this.overrides?.visitObject(proxy)
    }

    const result = []
    let entry

    while ((entry = proxy.nextEntry()) && !entry.done) {
      result.push(entry.value)
    }

    return Object.fromEntries(result)
  }

  visitIterable(access: IIterableAccess): T {
    if (isFunction(this.overrides?.visitIterable)) {
      return this.overrides?.visitIterable(access)
    }

    const result = []
    let element

    while ((element = access.nextElement())) {
      result.push(element)
    }

    return result as T
  }
}

export class ProxyDeserializer implements IDeserializer {
  private readonly deserializer: IDeserializer
  private readonly options?: SerdeOptions

  constructor(deserializer: IDeserializer, options?: SerdeOptions) {
    this.deserializer = deserializer
    this.options = options
  }

  deserializeAny<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeAny(new ProxyVisitor(visitor, this.options))
  }

  deserializeBoolean<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeBoolean(new ProxyVisitor(visitor, this.options))
  }

  deserializeNumber<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeNumber(new ProxyVisitor(visitor, this.options))
  }

  deserializeBigInt<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeBigInt(new ProxyVisitor(visitor, this.options))
  }

  deserializeString<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeString(new ProxyVisitor(visitor, this.options))
  }

  deserializeSymbol<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeSymbol(new ProxyVisitor(visitor, this.options))
  }

  deserializeNull<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeNull(new ProxyVisitor(visitor, this.options))
  }

  deserializeObject<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeObject(new ProxyVisitor(visitor, this.options))
  }

  deserializeIterable<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeIterable(new ProxyVisitor(visitor, this.options))
  }

  deserializeFunction<T, V extends IVisitor<T>>(visitor: V): T {
    return this.deserializer.deserializeFunction(new ProxyVisitor(visitor, this.options))
  }
}

