import { Deserialize } from './de'
import { Serialize } from './ser'
import { CaseConvention, convertCase } from './case'
import { isFunction, isIterable, isNumber, isString, Nullable } from './utils'

export interface RenameOptions {
  serialize?: string
  deserialize?: string
}

export interface RenameAllOptions {
  serialize?: CaseConvention
  deserialize?: CaseConvention
}

export interface ContainerOptions {
  rename?: RenameOptions | string
  renameAll?: RenameAllOptions | CaseConvention
  default?: () => any
  denyUnknownFields?: boolean
  tag?: string
  untagged?: boolean
}

interface Predicate {
  <T>(value: T): boolean
}

export interface SkipOptions {
  serialize?: Predicate | boolean
  deserialize?: Predicate | boolean
}
export interface PropertyOptions {
  alias?: string | string[]
  default?: () => any
  flatten?: boolean
  rename?: RenameOptions | string
  skip?: SkipOptions | Predicate | boolean
  serializeWith?: Serialize<any>
  deserializeWith?: Deserialize<any>
}

export const Stage = Object.freeze({
  Serialize: 0,
  Deserialize: 1
} as const)

export class PropertyMap extends Map<string, PropertyOptions> {
  private readonly options: SerdeOptions
  private readonly aliases: Map<string, string> = new Map()

  constructor(options: SerdeOptions, iterable?: Iterable<any>) {
    super(iterable)
    this.options = options
    this.buildAliasMap()
  }

  private setAliasesFromOptions(key: string, options: PropertyOptions) {
    this.aliases.set(key, key)

    if (options.alias != null) {
      this.addAlias(options.alias, key)
    }

    if (isString(options.rename)) {
      this.aliases.set(options.rename, key)
    } else if (isNumber(this.options.container.renameAll)) {
      this.aliases.set(this.options.applyCaseConvention(Stage.Deserialize, key), key)
    }
  }

  set(key: string, value: PropertyOptions): this {
    super.set(key, value)
    this.setAliasesFromOptions(key, value)
    return this
  }

  addAlias(alias: string | Iterable<string>, key: string) {
    if (isString(alias)) {
      this.aliases.set(alias, key)
    } else if (isIterable(alias)) {
      for (const a of alias) {
        this.addAlias(a, key)
      }
    }
  }

  getFieldFromAlias(alias: string) {
    return this.aliases.get(alias)
  }

  private buildAliasMap() {
    for (const [key, value] of this.entries()) {
      if (value.alias != null) {
        this.addAlias(value.alias, key)
      }
    }
  }
}

export type Stage = typeof Stage[keyof typeof Stage]

export class SerdeOptions {
  readonly target: Nullable<any>
  readonly container: ContainerOptions
  readonly properties: PropertyMap

  constructor(target: any, container: ContainerOptions = {}, properties: PropertyMap = new PropertyMap(this)) {
    this.target = target
    this.container = container
    this.properties = properties
  }

  private getClassName(stage: Stage, defaultName: string = this.target?.constructor?.name) {
    if (isString(this.container.rename)) {
      return this.container.rename
    } else if (stage === Stage.Serialize && isString(this.container.rename?.serialize)) {
      return this.container.rename.serialize
    } else if (stage === Stage.Deserialize && isString(this.container.rename?.deserialize)) {
      return this.container.rename.serialize
    } else {
      return defaultName
    }
  }

  getSerializedClassName(defaultName: string = this.target?.constructor?.name) {
    return this.getClassName(Stage.Serialize, defaultName)
  }

  getDeserializedClassName(defaultName: string = this.target?.constructor?.name) {
    return this.getClassName(Stage.Deserialize, defaultName)
  }

  getCaseConvention(stage: Stage) {
    if (isNumber(this.container.renameAll)) {
      if (stage === Stage.Serialize) {
        return this.container.renameAll
      } else {
        return CaseConvention.CamelCase
      }
    } else if (stage === Stage.Serialize && isNumber(this.container.renameAll?.serialize)) {
      return this.container.renameAll.serialize
    } else if (stage === Stage.Deserialize && isNumber(this.container.renameAll?.deserialize)) {
      return this.container.renameAll.deserialize
    }
  }

  applyCaseConvention(stage: Stage, property: string) {
    const convention = this.getCaseConvention(stage)
    return convention != null ? convertCase(property, convention) : property
  }

  getPropertyRename(stage: Stage, property: string) {
    const options = this.properties.get(property)

    if (options == null) {
      return
    } else if (isString(options.rename)) {
      return options.rename
    } else if (stage == Stage.Serialize && isString(options.rename?.serialize)) {
      return options.rename.serialize
    } else if (stage == Stage.Deserialize && isString(options.rename?.deserialize)) {
      return options.rename.deserialize
    }
  }

  getSerializationPropertyName(property: string) {
    return this.getPropertyRename(Stage.Serialize, property) ||
      this.applyCaseConvention(Stage.Serialize, property)
  }

  getNameFromAlias(alias: string) {
    return this.properties.getFieldFromAlias(alias)
  }

  getDeserializationPropertyName(property: string) {
    return this.getNameFromAlias(property) ||
      this.applyCaseConvention(Stage.Deserialize, property)
  }

  shouldSkip(stage: Stage, property: string, value: any) {
    const options = this.properties.get(property)
    if (options == null) {
      return false
    } else if (typeof options.skip === 'boolean') {
      return options.skip
    } else if (isFunction(options.skip)) {
      return options.skip(value)
    } else if (stage === Stage.Serialize && typeof options.skip?.serialize === 'boolean') {
      return options.skip.serialize
    } else if (stage === Stage.Serialize && isFunction(options.skip?.serialize)) {
      return options.skip.serialize(value)
    } else if (stage === Stage.Deserialize && typeof options.skip?.deserialize === 'boolean') {
      return options.skip.deserialize
    } else if (stage === Stage.Deserialize && isFunction(options.skip?.deserialize)) {
      return options.skip.deserialize(value)
    } else {
      return false
    }
  }

  shouldSerialize(property: string, value: any) {
    return !this.shouldSkip(Stage.Serialize, property, value)
  }

  shouldDeserialize(property: string, value: any) {
    return !this.shouldSkip(Stage.Deserialize, property, value)
  }

  defaultFor(property: string) {
    const options = this.properties.get(property)
    if (options != null && isFunction(options.default)) {
      return options.default()
    } else if (isFunction(this.container.default)) {
      return this.container.default()
    }
  }
}

